/*
** Module   : ImageGuiImpl.java
** Abstract : Implementation of a IMAGE widget for the GUI driver.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.  Not yet fully tested/complete.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 CA  20141021 Fixed image drawing and loading: tested only for the case when is used as a 
**                  window icon.
** 006 MAG 20141029 Added clipping rectangle when drawing. 
** 007 HC  20141021 Implemented window move/position support.
** 008 EVL 20141204 Adding image reloading capability.
** 009 EVL 20141208 Adding image widget initialization code.
** 010 EVL 20150108 Differentiate small icon from large one flag.  Rework for icon/image
**                  implementation.
** 011 EVL 20150123 Handle deferred image loading.
** 012 EVL 20150203 Implementing image offset capability and size phrase.  Change the approach for
**                  getting image coordinates.
** 013 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
*/

package com.goldencode.p2j.ui.client.gui;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import java.awt.image.*;
import javax.imageio.*;
import javax.imageio.stream.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of a IMAGE widget for the GUI driver.
 */
public class ImageGuiImpl
extends Image<GuiOutputManager>
{
   /** logger */
   protected static final Logger LOG = LogHelper.getLogger(ImageGuiImpl.class.getName());

   /** The image object. */
   protected java.awt.Image img;
   
   /** The graphics driver. */
   private GuiDriver gd;
   
   /** To keep track if this is the window icon or not. */
   private boolean isIcon = false;
   
   /**
    * Create a new image widget.
    * 
    * @param    id
    *           The widget's ID.
    */
   public ImageGuiImpl(WidgetId id)
   {
      super(id);

      this.gd = (GuiDriver) OutputManager.getDriver();
   }
   
   /**
    * Create a new image widget.
    * 
    * @param    id
    *           The widget's ID.
    * @param    isIcon
    *           Flag indicating the image is the window title bar icoon.
    */
   public ImageGuiImpl(WidgetId id, boolean isIcon)
   {
      this(id);

      this.isIcon = isIcon;
   }
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ImageGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ImageConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig cfg)
   {
      super.initialize(id, cfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // check if we need to preload image
      String imgSrcFile = ((ImageConfig) cfg).image;
      // filename exists, load deferred image
      if (imgSrcFile != null && imgSrcFile.length() > 0 && img == null)
      {
         int thisId = config().id.getId();
         int type = ButtonConfig.IMAGE_IMG;

         // first try to search for images loaded from application jar
         byte[] imgData = ImageHelper.getPendingImageData(thisId, type);
         if (imgData != null)
         {
            setImage(imgData);
         }
         else
         {
            String loadName = ImageHelper.getPendingImageName(thisId, type);
            if (loadName != null && loadName.length() > 0)
            {
               loadImage(loadName);
            }
         }
      }
   }

   /**
    * Draw this image
    */
   @Override
   public void draw()
   {
      
      NativePoint origin = physicalLocation();
      final NativeDimension d = physicalDimension();
      final NativeRectangle clip = clipRectangle(origin, d);
      
      gd.draw(origin, clip, new Runnable()
      {
         @Override
         public void run()
         {
            if (!drawImage())
            {
               gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
               gd.fillRect(0, 0, d.width, d.height);
            }
         }
      });
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      ImageConfig icfg = (ImageConfig) config();
      // here we need the actual image height
      int px = img == null ? 0 : icfg.heightPixels != 0 ? icfg.heightPixels : img.getHeight(null);
      return screen().coordinates().heightFromNative(px);
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      ImageConfig icfg = (ImageConfig) config();
      // here we need the actual image width
      int px = img == null ? 0 : icfg.widthPixels != 0 ? icfg.widthPixels : img.getWidth(null);
      return screen().coordinates().widthFromNative(px);
   }

   /**
    * Sets the new image for given image widget on the client side.
    * 
    * @param    byteImg
    *           A new image array data.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean setImage(byte[] byteImg)
   {
      // the new image candidate
      java.awt.Image imgNew = null;

      // try creating new image
      try
      {
         imgNew = ImageIO.read(new ByteArrayInputStream(byteImg));
      }
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not load image from byte array...", e);
         }
         imgNew = null;
      }
      
      // image created, replace the old one
      if (imgNew != null)
      {
         img = imgNew;
         // repaint changed image
         ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
         {
            @Override
            public void run()
            {
               repaint();
            }
         });
         return true;
      }
      else
      {
         return false;
      }
   }
   
   /**
    * Loads the new image for given image widget from the client side filesystem.
    * 
    * @param    fullName
    *           The fully qualified file name for the image to be loaded.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public boolean loadImage(String fullName)
   {
      InputStream is = null;
      byte[] imageData = null;
         
      // find the image on local file system
      try
      {
         is = new FileInputStream(fullName);
      }
      catch (FileNotFoundException ex)
      {
         is = null;
      }
         
      // the file is found, try to read
      if (is != null)
      {
         imageData = UiUtils.getData(is);
      }
         
      // if having new image data - set it
      if (imageData != null)
      {
         // remember if this is an icon or not
         isIcon = fullName.indexOf(".ico") != -1;
         return setImage(imageData);
      }
      else
      {
         return false;
      }
   }
   
   /**
    * Get internal image native width in pixels.
    * 
    * @return  internal image width.
    */
   protected int getImageWidth()
   {
      if (img != null)
      {
         ImageConfig icfg = (ImageConfig) config();
         return icfg.widthPixels > 0 ?
            Math.min(((BufferedImage)img).getWidth(), icfg.widthPixels) :
            ((BufferedImage)img).getWidth();
      }
      else
      {
         return screen().coordinates().columnToNative(width());
      }
   }
   
   /**
    * Get internal image native height in pixels.
    * 
    * @return  internal image height.
    */
   protected int getImageHeight()
   {
      if (img != null)
      {
         ImageConfig icfg = (ImageConfig) config();
         return icfg.heightPixels > 0 ?
            Math.min(((BufferedImage)img).getHeight(), icfg.heightPixels) :
            ((BufferedImage)img).getHeight();
      }
      else
      {
         return screen().coordinates().rowToNative(height());
      }
   }
   
   /**
    * Draw this image.
    */
   protected boolean drawImage()
   {
      if (img != null)
      {
         CoordinatesConversion c = screen().coordinates();
         
         // get the current image config
         ImageConfig icfg = (ImageConfig) config();

         // the width and height in pixels or chars overridden by size phrase has higher priority
         // than natural image size itself
         int w = icfg.widthPixels > 0 ? icfg.widthPixels : icfg.widthChars > 0 ?
                    c.widthToNative(icfg.widthChars) : c.widthToNative(width());
         int h = icfg.heightPixels > 0 ? icfg.heightPixels : icfg.heightChars > 0 ?
                    c.heightToNative(icfg.heightChars) : c.heightToNative(height());

         // compute final image dimensions
         int iw = 0;
         int ih = 0;
         if (icfg.stretchToFit && !isIcon)
         {
            // need to fit the image to display area size
            iw = w;
            ih = h;
         }
         else
         {
            // take the final image size as minimum from requested and real image sizes
            iw = Math.min(((BufferedImage)img).getWidth(), w);
            ih = Math.min(((BufferedImage)img).getHeight(), h);
         }
         
         // recalculate image offsets that can be changed
         int offsetX = -1;
         int offsetY = -1;
         // check if we have the offset variables changed
         if (icfg.offsetX != -1)
         {
            offsetX = icfg.offsetX;
         }
         else if (icfg.offsetCol != 0.0)
         {
            offsetX = screen().coordinates().columnToNative(icfg.offsetCol);
         }
         else if (icfg.widthPixels != 0 || icfg.widthChars != 0.0)
         {
            offsetX = 0;
         }
         // Y coordinate
         if (icfg.offsetY != -1)
         {
            offsetY = icfg.offsetY;
         }
         else if (icfg.offsetRow != 0.0)
         {
            offsetY = screen().coordinates().rowToNative(icfg.offsetRow);
         }
         else if (icfg.heightPixels != 0 || icfg.heightChars != 0.0)
         {
            offsetY = 0;
         }
         // render image
         gd.drawImage(img, 0, 0, iw, ih, offsetX, offsetY, icfg.transparent,
                      icfg.stretchToFit && !isIcon, icfg.retainShape, icfg.convert3D);
      }
      
      return img != null;
   }
}
